@@ -37,6 +37,7 @@ from pydantic.utils import Representation
 from redis.client import Pipeline
 from ulid import ULID
 
+from ..checks import has_redis_json, has_redisearch
 from ..connections import get_redis_connection
 from .encoders import jsonable_encoder
 from .render_tree import render_tree
@@ -121,27 +122,18 @@ def validate_model_fields(model: Type["RedisModel"], field_values: Dict[str, Any
             )
 
 
-class ExpressionProtocol(Protocol):
-    op: Operators
-    left: ExpressionOrModelField
-    right: ExpressionOrModelField
-
-    def __invert__(self) -> "Expression":
-        pass
-
-    def __and__(self, other: ExpressionOrModelField):
-        pass
-
-    def __or__(self, other: ExpressionOrModelField):
-        pass
-
-    @property
-    def name(self) -> str:
-        raise NotImplementedError
-
-    @property
-    def tree(self) -> str:
-        raise NotImplementedError
+def decode_redis_value(
+    obj: Union[List[bytes], Dict[bytes, bytes], bytes], encoding: str
+) -> Union[List[str], Dict[str, str], str]:
+    """Decode a binary-encoded Redis hash into the specified encoding."""
+    if isinstance(obj, list):
+        return [v.decode(encoding) for v in obj]
+    if isinstance(obj, dict):
+        return {
+            key.decode(encoding): value.decode(encoding) for key, value in obj.items()
+        }
+    elif isinstance(obj, bytes):
+        return obj.decode(encoding)
 
 
 @dataclasses.dataclass
@@ -317,6 +309,13 @@ class FindQuery:
         page_size: int = DEFAULT_PAGE_SIZE,
         sort_fields: Optional[List[str]] = None,
     ):
+        if not has_redisearch(model.db()):
+            raise RedisModelError(
+                "Your Redis instance does not have either the RediSearch module "
+                "or RedisJSON module installed. Querying requires that your Redis "
+                "instance has one of these modules installed."
+            )
+
         self.expressions = expressions
         self.model = model
         self.offset = offset
@@ -330,8 +329,8 @@ class FindQuery:
 
         self._expression = None
         self._query: Optional[str] = None
-        self._pagination: list[str] = []
-        self._model_cache: list[RedisModel] = []
+        self._pagination: List[str] = []
+        self._model_cache: List[RedisModel] = []
 
     def dict(self) -> Dict[str, Any]:
         return dict(
@@ -792,7 +791,7 @@ class FindQuery:
         return query.execute()[0]
 
 
-class PrimaryKeyCreator(Protocol):
+class PrimaryKeyCreator(abc.ABC):
     def create_pk(self, *args, **kwargs) -> str:
         """Create a new primary key"""
 
@@ -919,6 +918,7 @@ class MetaProtocol(Protocol):
     index_name: str
     abstract: bool
     embedded: bool
+    encoding: str
 
 
 @dataclasses.dataclass
@@ -938,6 +938,7 @@ class DefaultMeta:
     index_name: Optional[str] = None
     abstract: Optional[bool] = False
     embedded: Optional[bool] = False
+    encoding: str = "utf-8"
 
 
 class ModelMeta(ModelMetaclass):
@@ -1007,10 +1008,13 @@ class ModelMeta(ModelMetaclass):
             new_class._meta.database = getattr(
                 base_meta, "database", get_redis_connection()
             )
+        if not getattr(new_class._meta, "encoding", None):
+            new_class._meta.encoding = getattr(base_meta, "encoding")
         if not getattr(new_class._meta, "primary_key_creator_cls", None):
             new_class._meta.primary_key_creator_cls = getattr(
                 base_meta, "primary_key_creator_cls", UlidPrimaryKey
             )
+        # TODO: Configurable key separate, defaults to ":"
         if not getattr(new_class._meta, "index_name", None):
             new_class._meta.index_name = (
                 f"{new_class._meta.global_key_prefix}:"
@@ -1059,7 +1063,7 @@ class RedisModel(BaseModel, abc.ABC, metaclass=ModelMeta):
     def save(self, pipeline: Optional[Pipeline] = None) -> "RedisModel":
         raise NotImplementedError
 
-    @validator("pk", always=True)
+    @validator("pk", always=True, allow_reuse=True)
     def validate_pk(cls, v):
         if not v:
             v = cls._meta.primary_key_creator_cls().create_pk()
@@ -1200,12 +1204,34 @@ class HashModel(RedisModel, abc.ABC):
         db.hset(self.key(), mapping=document)
         return self
 
+    @classmethod
+    def all_pks(cls):
+        key_prefix = cls.make_key(cls._meta.primary_key_pattern.format(pk=""))
+        # TODO: We assume the key ends with the default separator, ":" -- when
+        #  we make the separator configurable, we need to update this as well.
+        #  ... And probably lots of other places ...
+        return (
+            key.split(":")[-1]
+            for key in cls.db().scan_iter(f"{key_prefix}*", _type="HASH")
+        )
+
     @classmethod
     def get(cls, pk: Any) -> "HashModel":
         document = cls.db().hgetall(cls.make_primary_key(pk))
         if not document:
             raise NotFoundError
-        return cls.parse_obj(document)
+        try:
+            result = cls.parse_obj(document)
+        except TypeError as e:
+            log.warning(
+                f'Could not parse Redis response. Error was: "{e}". Probably, the '
+                "connection is not set to decode responses from bytes. "
+                "Attempting to decode response using the encoding set on "
+                f"model class ({cls.__class__}. Encoding: {cls.Meta.encoding}."
+            )
+            document = decode_redis_value(document, cls.Meta.encoding)
+            result = cls.parse_obj(document)
+        return result
 
     @classmethod
     @no_type_check
@@ -1316,6 +1342,11 @@ class HashModel(RedisModel, abc.ABC):
 
 class JsonModel(RedisModel, abc.ABC):
     def __init_subclass__(cls, **kwargs):
+        if not has_redis_json(cls.db()):
+            log.error(
+                "Your Redis instance does not have the RedisJson module "
+                "loaded. JsonModel depends on RedisJson."
+            )
         # Generate the RediSearch schema once to validate fields.
         cls.redisearch_schema()
 
